##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

'''$Id: db.py,v 1.13 2002/05/15 21:14:23 matt Exp $'''
__version__='$Revision: 1.13 $'[11:-2]

from Globals import INSTANCE_HOME
import os
import re
import apsw
# from DateTime import DateTime
from Shared.DC.ZRDB.TM import TM

import string, sys
from string import strip, split, find
from time import time
from SQLite3Utils import SQLite3
failures=0
calls=0
last_call_time=time()

IMPLICITLOBS=1      # Turn this off to get old LOB behavior

class DB(TM):

    _p_oid=_p_changed=_registered=None

   # Database_Connection=apsw.Connection
    Database_Error=apsw.Error

    def __init__(self,connection_string):
        self.connection_string=connection_string
        self.dbpath = os.path.join(INSTANCE_HOME,'var','sqlite3',
				connection_string) 
        self.Connect()

    def Connect(self) :
        self.db=SQLite3(self.dbpath)

    def str(self,v, StringType=type('')):
        if v is None: return ''
        r=str(v)
        if r[-1:]=='L' and type(v) is not StringType: r=r[:-1]
        return r

    def _begin(self, *ignored):
        self.check_threadingError()
        self.db.BEGIN()

    def _finish(self, *ignored):
        self.check_threadingError()
        self.db.COMMIT()

    def _abort(self, *ignored):
        self.check_threadingError()
        self.db.ROLLBACK()

    def tables(self, rdb=0, _care=('table', 'view')):
        self.check_threadingError()
        c = self.db.cursor
        try :
            items,result = self.query("select * from sqlite_master")
            r = []
            a=r.append
            for i in result :
                if len(_care)==0 or i[0].lower() in _care:
                    a({'TABLE_NAME': i[1].lower(), 'TABLE_TYPE': i[0].lower()})
            return r
        except :
            return []

    type_re = re.compile("([^(]*)(\(([0-9]+)(,[0-9]+)?\))?")
    def _typeparser(self,typ) :
        m = type_re.match(typ)
        return (m.group(1),m.group(3),m.group(5))

    def columns(self, table_name):
        self.check_threadingError()
        c=self.db.cursor
        try: r=c.execute('select * from %s' % table_name)
        except: return ()
        desc=r.getdescription()
        r=[]
        a=r.append
        for name, typ1 in desc:
            typ,length,dec = self._typeparser(typ1)
            a({ 'Name': name.lower(),
                'Type': typ.lower(),
                'Precision': length,
                'Scale': dec,
                'Nullable': '',
                })

        return r

    def query(self,query_string, max_rows=9999999):
        self.check_threadingError()
        global failures, calls, last_call_time
        items = []
        calls=calls+1
        desc=None
        result=()
        self._register()
        c=self.db.cursor            
	if c==None :
            return items,result
        try:
            r=c.execute(query_string) # Returns 1 on SELECT
            try:
                desc = r.getdescription()
                for i in desc :
                    items.append({'name':i[0],'type':i[1]})
            except :
                desc = None
            if desc is not None:
                if type(result) is not type([]) :
                    result=list(result)
                try :
                    rec = r.next()
                    while rec :
                        result.append(rec)
                        rec = r.next()
                except StopIteration :
                    pass
            failures=0
            last_call_time=time()
        except self.Database_Error, mess:
            failures=failures+1
            if (failures > 1000 or time()-last_call_time > 600):
                # Hm. maybe the db is hosed.  Let's try once to restart it.
                failures=0
                last_call_time=time()
                self.db.CLOSE()
                self.db._connection()
                return self.query(query_string, max_rows)
	    else: raise sys.exc_type, sys.exc_value, sys.exc_traceback

        if desc is None: return (),()

        return items, result

    def check_threadingError(self) :
        try :
            self.db.cursor.execute('select * from sqlite_master')
            self.db._eatCursorNext()
        except apsw.ThreadingViolationError:
            self.db.CLOSE()
            self.db._connection()

